import os
from pathlib import Path

import click
from qualibrate_config.vars import (
    DEFAULT_CONFIG_FILENAME,
    QUALIBRATE_PATH,
)

from qualibrate_app.config import CONFIG_PATH_ENV_NAME


@click.command(name="start")
@click.option(
    "--config-path",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        path_type=Path,
    ),
    default=QUALIBRATE_PATH / DEFAULT_CONFIG_FILENAME,
    help="Path to `config.toml` file",
    show_default=True,
)
@click.option(
    "--reload", is_flag=True, hidden=True
)  # env QUALIBRATE_START_RELOAD
@click.option(
    "--port",
    type=int,
    default=8001,
    show_default=True,
    help="Application will be started on the given port",
)  # env QUALIBRATE_START_PORT
def start_command(config_path: Path, port: int, reload: bool) -> None:
    os.environ[CONFIG_PATH_ENV_NAME] = str(config_path)

    from qualibrate_app.app import main as app_main

    app_main(port=port, reload=reload)



from pathlib import Path
from typing import Callable, Optional

from qualibrate_app.api.core.types import IdType
from qualibrate_app.api.core.utils.path.node import NodePath
from qualibrate_app.api.core.utils.singleton import Singleton
from qualibrate_app.api.exceptions.classes.storage import QFileNotFoundException

__all__ = ["default_node_path_solver", "IdToLocalPath", "NodePathSolverType"]


NodePathSolverType = Callable[[IdType, Path], Optional[NodePath]]


def default_node_path_solver(id: IdType, base_path: Path) -> Optional[NodePath]:
    node_path = next(base_path.glob(f"*/#{id}_*"), None)
    return NodePath(node_path) if node_path is not None else None


class _IdToProjectLocalPath:
    def __init__(self, project_name: str, project_path: Path) -> None:
        self._mapping: dict[IdType, NodePath] = {}
        self._project_name = project_name
        self._project_path = project_path

    def get(
        self,
        id: IdType,
        solver: NodePathSolverType = default_node_path_solver,
    ) -> Optional[NodePath]:
        if id in self._mapping:
            return self._mapping[id]
        path = solver(id, self._project_path)
        if path is None:
            return None
        self._mapping[id] = path
        return path


class IdToLocalPath(metaclass=Singleton):
    def __init__(self) -> None:
        self._project_to_path: dict[str, _IdToProjectLocalPath] = {}

    def get(
        self,
        project: str,
        id: IdType,
        project_path: Path,
        solver: NodePathSolverType = default_node_path_solver,
    ) -> Optional[NodePath]:
        if project not in self._project_to_path:
            self._project_to_path[project] = _IdToProjectLocalPath(
                project, project_path
            )
        p2p = self._project_to_path[project]
        path = p2p.get(id, solver)
        return path

    def get_or_raise(
        self,
        project: str,
        id: IdType,
        project_path: Path,
        solver: NodePathSolverType = default_node_path_solver,
    ) -> NodePath:
        path = self.get(project, id, project_path, solver)
        if path is None:
            raise QFileNotFoundException(
                f"Node {id} of project '{project}' not found"
            )
        return path



# COPIED FROM qualibrate-core/qualibrate/utils/types_parsing.py

import sys
from collections.abc import Mapping
from typing import Any, Optional, Union

if sys.version_info >= (3, 10):
    from types import NoneType
else:
    NoneType = type(None)

NOT_NONE_BASIC_TYPES = Union[bool, int, float, str]
BASIC_TYPES = Union[NOT_NONE_BASIC_TYPES, NoneType]
LIST_TYPES = Union[list[bool], list[int], list[float], list[str]]
VALUE_TYPES_WITHOUT_REC = Union[BASIC_TYPES, LIST_TYPES]
INPUT_CONVERSION_TYPE = Union[
    VALUE_TYPES_WITHOUT_REC, dict[str, "INPUT_CONVERSION_TYPE"]
]


def parse_bool(value: VALUE_TYPES_WITHOUT_REC) -> VALUE_TYPES_WITHOUT_REC:
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        if value.lower() == "true":
            return True
        if value.lower() == "false":
            return False
        return value
    if isinstance(value, int):
        return value != 0
    if isinstance(value, float):
        return abs(value) > 1e-9
    return value


def parse_int(value: VALUE_TYPES_WITHOUT_REC) -> VALUE_TYPES_WITHOUT_REC:
    if isinstance(value, bool):
        return int(value)
    if isinstance(value, int):
        return value
    if isinstance(value, str):
        if value.isdigit() or (
            len(value) > 1 and value[0] == "-" and value[1:].isdigit()
        ):
            return int(value)
        return value
    if isinstance(value, float):
        if value.is_integer():
            return int(value)
        return value
    return value


def parse_float(value: VALUE_TYPES_WITHOUT_REC) -> VALUE_TYPES_WITHOUT_REC:
    if isinstance(value, float):
        return value
    if isinstance(value, int):
        return float(value)
    if isinstance(value, str):
        try:
            return float(value)
        except ValueError:
            return value
    return value


def parse_str(value: VALUE_TYPES_WITHOUT_REC) -> VALUE_TYPES_WITHOUT_REC:
    if not isinstance(value, str):
        return str(value)
    if len(value) < 3 or value[0] != value[-1] or value[0] != '"':
        return value
    # remove wrapping quotes passed from FE
    return value[1:-1]


def parse_none(value: VALUE_TYPES_WITHOUT_REC) -> VALUE_TYPES_WITHOUT_REC:
    if value is None or (isinstance(value, str) and not value):
        return None
    return value


BASIC_PARSERS = {
    int: parse_int,
    float: parse_float,
    bool: parse_bool,
    str: parse_str,
    NoneType: parse_none,
}

STR_TO_TYPE: Mapping[str, type[BASIC_TYPES]] = {
    "integer": int,
    "number": float,
    "boolean": bool,
    "string": str,
    "null": NoneType,
}

TYPE_TO_STR = {v: k for k, v in STR_TO_TYPE.items()}


def parse_typed_list(
    value: list[Any], item_type: type[BASIC_TYPES]
) -> LIST_TYPES:
    if len(value) == 0:
        return value
    parser = BASIC_PARSERS.get(item_type)
    if parser is None:
        return value
    try:
        return list(map(parser, value))  # type: ignore
    except ValueError:
        return value


def _parse_list(
    value: list[Any],
    item_type: Optional[type[BASIC_TYPES]],
) -> VALUE_TYPES_WITHOUT_REC:
    if item_type is None:
        return value
    return parse_typed_list(value, item_type)


def parse_list(
    value: VALUE_TYPES_WITHOUT_REC,
    item_type: Optional[type[BASIC_TYPES]],
) -> VALUE_TYPES_WITHOUT_REC:
    if isinstance(value, list):
        return _parse_list(value, item_type)
    if isinstance(value, str):
        stripped = value.strip()
        if stripped.startswith("[") and stripped.endswith("]"):
            stripped = stripped[1:-1]
        splitted = list(map(str.strip, stripped.split(",")))
        return _parse_list(splitted, item_type)
    return value


def types_conversion(value: Any, expected_type: Mapping[str, Any]) -> Any:
    if isinstance(value, Mapping):
        # has sub items
        new = {}
        for k, v in value.items():
            if k in expected_type:
                new[k] = types_conversion(v, expected_type[k])
            else:
                new[k] = v
        return new
    if "anyOf" in expected_type:
        # suppose that only `Type | None` is possible
        none = parse_none(value)
        if none is None:
            return None
        expected_type_ = dict(expected_type)
        expected_type_["type"] = expected_type_.pop("anyOf")[0]["type"]
        return types_conversion(value, expected_type_)
    if "type" in expected_type:
        if expected_type.get("type") == "array":
            # array
            item_type: Optional[type[BASIC_TYPES]] = (
                STR_TO_TYPE.get(expected_type["items"]["type"])
                if "items" in expected_type
                else None
            )
            return parse_list(value, item_type)
        if expected_type.get("type") in STR_TO_TYPE:
            expected = STR_TO_TYPE[expected_type["type"]]
            parser = BASIC_PARSERS[expected]
            return parser(value)
    return value



from abc import ABC

from qualibrate_config.models import (
    JsonTimelineDBRemoteServiceConfig,
    QualibrateAppConfig,
    QualibrateConfig,
)


class DomainWithConfigBase(ABC):  # noqa: B024
    def __init__(self, settings: QualibrateConfig):
        self._settings = settings

    @property
    def app_config(self) -> QualibrateAppConfig:
        if self._settings.app is None:
            raise RuntimeError("Qualibrate app config not set")
        return self._settings.app

    @property
    def timeline_db_config(self) -> JsonTimelineDBRemoteServiceConfig:
        app = self.app_config
        if app.timeline_db is None:
            raise RuntimeError("Timeline DB of Qualibrate app config not set")
        return app.timeline_db



from collections.abc import Sequence
from datetime import datetime
from typing import Optional

from qualibrate_config.models import QualibrateConfig

from qualibrate_app.api.core.domain.bases.branch import (
    BranchBase,
    BranchLoadType,
)
from qualibrate_app.api.core.domain.bases.node import NodeBase, NodeLoadType
from qualibrate_app.api.core.domain.bases.snapshot import (
    SnapshotBase,
    SnapshotLoadType,
)
from qualibrate_app.api.core.domain.timeline_db.node import NodeTimelineDb
from qualibrate_app.api.core.domain.timeline_db.snapshot import (
    SnapshotTimelineDb,
)
from qualibrate_app.api.core.types import DocumentType, IdType
from qualibrate_app.api.core.utils.request_utils import request_with_db
from qualibrate_app.api.exceptions.classes.timeline_db import QJsonDbException

__all__ = ["BranchTimelineDb"]


class BranchTimelineDb(BranchBase):
    def __init__(
        self,
        name: str,
        content: Optional[DocumentType] = None,
        *,
        settings: QualibrateConfig,
    ):
        super().__init__(name, content, settings=settings)

    @property
    def created_at(self) -> Optional[datetime]:
        if "created_at" not in self.content:
            return None
        return datetime.fromisoformat(
            str(self.content.get("created_at"))
        ).astimezone()

    def load(self, load_type: BranchLoadType) -> None:
        if self._load_type == BranchLoadType.Full:
            return
        timeline_db_config = self.timeline_db_config
        result = request_with_db(
            f"branch/{self._name}/",
            db_name=self._settings.project,
            host=timeline_db_config.address_with_root,
            timeout=timeline_db_config.timeout,
        )
        no_branch_ex = QJsonDbException("Branch data wasn't retrieved.")
        if result.status_code != 200:
            raise no_branch_ex
        content = result.json()
        if content is None:
            raise no_branch_ex
        self.content.update(content)
        self._load_type = BranchLoadType.Full

    def get_snapshot(self, id: Optional[IdType] = None) -> SnapshotBase:
        if id is None:
            latest = self.get_latest_snapshots(1)
            if len(latest) != 1:
                raise QJsonDbException("Can't load latest snapshot of branch")
            return latest[1][0]
        timeline_db_config = self.timeline_db_config
        res = request_with_db(
            f"branch/{self.name}/is_snapshot_belong",
            params={"snapshot_id": id},
            host=timeline_db_config.address_with_root,
            db_name=self._settings.project,
            timeout=timeline_db_config.timeout,
        )
        snapshot_belonged_to_branch = bool(res.json())
        if not snapshot_belonged_to_branch:
            raise QJsonDbException("Snapshot doesn't belong to branch.")
        snapshot = SnapshotTimelineDb(id=id, settings=self._settings)
        snapshot.load(SnapshotLoadType.Metadata)
        return snapshot

    def get_node(self, id: Optional[IdType] = None) -> NodeBase:
        if id is None:
            latest = self.get_latest_nodes(1)
            if len(latest) != 1:
                raise QJsonDbException("Can't load latest node of branch")
            return latest[1][0]
        timeline_db_config = self.timeline_db_config
        res = request_with_db(
            f"branch/{self.name}/is_snapshot_belong",
            params={"snapshot_id": id},
            host=timeline_db_config.address_with_root,
            db_name=self._settings.project,
            timeout=timeline_db_config.timeout,
        )
        snapshot_belonged_to_branch = bool(res.json())
        if not snapshot_belonged_to_branch:
            raise QJsonDbException("Node snapshot doesn't belong to branch.")
        node = NodeTimelineDb(node_id=id, settings=self._settings)
        node.load(NodeLoadType.Full)
        return node

    def _get_remote_snapshots(
        self,
        metadata: bool,
        page: int,
        per_page: int,
        reverse: bool,
    ) -> tuple[int, Sequence[DocumentType]]:
        timeline_db_config = self.timeline_db_config
        result = request_with_db(
            f"branch/{self._name}/history",
            params={
                "metadata": metadata,
                "page": page,
                "per_page": per_page,
                "reverse": reverse,
            },
            host=timeline_db_config.address_with_root,
            db_name=self._settings.project,
            timeout=timeline_db_config.timeout,
        )
        if result.status_code != 200:
            raise QJsonDbException("Branch history wasn't retrieved.")
        snapshots = result.json()
        if not isinstance(snapshots, Sequence):
            raise QJsonDbException("Branch history wasn't retrieved.")
        # TODO: func for parse all paged items
        parsed = result.json()
        return parsed["total"], list(parsed["items"])

    def get_latest_snapshots(
        self,
        page: int = 0,
        per_page: int = 50,
        reverse: bool = False,
    ) -> tuple[int, list[SnapshotBase]]:
        """Retrieve last num_snapshots from this branch"""
        total, snapshots = self._get_remote_snapshots(
            True, page, per_page, reverse
        )
        return total, [
            SnapshotTimelineDb(
                id=snapshot["id"], content=snapshot, settings=self._settings
            )
            for snapshot in snapshots
        ]

    def get_latest_nodes(
        self,
        page: int = 0,
        per_page: int = 50,
        reverse: bool = False,
    ) -> tuple[int, list[NodeBase]]:
        """Retrieve last num_snapshots from this branch"""
        total, snapshots = self._get_remote_snapshots(
            False, page, per_page, reverse
        )
        return total, [
            NodeTimelineDb(
                node_id=snapshot["id"],
                snapshot_content=snapshot,
                settings=self._settings,
            )
            for snapshot in snapshots
        ]



from collections.abc import Sequence
from pathlib import Path
from typing import Annotated

from fastapi import APIRouter, Depends
from qualibrate_config.models import QualibrateConfig, StorageType

from qualibrate_app.api.core.domain.bases.project import ProjectsManagerBase
from qualibrate_app.api.core.domain.local_storage.project import (
    ProjectsManagerLocalStorage,
)
from qualibrate_app.api.core.domain.timeline_db.project import (
    ProjectsManagerTimelineDb,
)
from qualibrate_app.api.core.models.project import Project
from qualibrate_app.config import (
    get_config_path,
    get_settings,
)

project_router = APIRouter(prefix="/projects", tags=["project"])


def _get_projects_manager(
    settings: Annotated[QualibrateConfig, Depends(get_settings)],
    config_path: Annotated[Path, Depends(get_config_path)],
) -> ProjectsManagerBase:
    project_types = {
        StorageType.local_storage: ProjectsManagerLocalStorage,
        StorageType.timeline_db: ProjectsManagerTimelineDb,
    }
    return project_types[settings.storage.type](
        settings=settings, config_path=config_path
    )


@project_router.get("/list")
def get_projects_list(
    projects_manager: Annotated[
        ProjectsManagerBase, Depends(_get_projects_manager)
    ],
) -> Sequence[Project]:
    return projects_manager.list()


@project_router.post("/create")
def create_project(
    project_name: str,
    projects_manager: Annotated[
        ProjectsManagerBase, Depends(_get_projects_manager)
    ],
) -> str:
    return projects_manager.create(project_name)


@project_router.get("/active")
def get_active_project(
    projects_manager: Annotated[
        ProjectsManagerBase, Depends(_get_projects_manager)
    ],
) -> str:
    return projects_manager.project


@project_router.post("/active")
def set_active_project(
    active_project: str,
    projects_manager: Annotated[
        ProjectsManagerBase, Depends(_get_projects_manager)
    ],
) -> str:
    projects_manager.project = active_project
    return active_project



from collections.abc import Mapping
from functools import partial
from typing import Any, Callable, Optional, Union
from urllib.parse import urljoin

import requests
from fastapi import HTTPException
from pydantic import HttpUrl
from qualibrate_config.models import (
    QualibrateConfig,
    QualibrateRunnerRemoteServiceConfig,
)

HTTPException422 = partial(HTTPException, status_code=422)


def get_runner_config(
    config: QualibrateConfig,
) -> QualibrateRunnerRemoteServiceConfig:
    if config.runner is None:
        raise RuntimeError("Qualibrate runner remote service is not configured")
    return config.runner


def request_with_db(
    path: str,
    *,
    params: Optional[Mapping[str, Any]] = None,
    db_name: str,
    timeout: float,
    method: Callable[..., requests.Response] = requests.get,
    host: Union[HttpUrl, str],
    **kwargs: Any,
) -> requests.Response:
    if params is None:
        params = {"db_name": db_name}
    else:
        params = dict(params)
        params["db_name"] = db_name
    result = method(
        urljoin(str(host), path),
        params=params,
        timeout=timeout,
        **kwargs,
    )
    return result



from datetime import date, datetime
from functools import cached_property

from qualibrate_app.api.core.utils.path import ConcretePath

__all__ = ["NodesDatePath"]


class NodesDatePath(ConcretePath):
    @cached_property
    def date(self) -> date:
        return self.datetime.date()

    @cached_property
    def datetime(self) -> datetime:
        return datetime.strptime(self.stem, "%Y-%m-%d")



import contextlib
from collections.abc import Mapping
from typing import Annotated, Any, Optional, Union, cast
from urllib.parse import urljoin

import requests
from fastapi import APIRouter, Body, Cookie, Depends, Path, Query
from qualibrate_config.models import QualibrateConfig, StorageType

from qualibrate_app.api.core.domain.bases.snapshot import (
    SnapshotBase,
    SnapshotLoadType,
)
from qualibrate_app.api.core.domain.local_storage.snapshot import (
    SnapshotLocalStorage,
    logger,
)
from qualibrate_app.api.core.domain.timeline_db.snapshot import (
    SnapshotTimelineDb,
)
from qualibrate_app.api.core.models.paged import PagedCollection
from qualibrate_app.api.core.models.snapshot import (
    SimplifiedSnapshotWithMetadata,
)
from qualibrate_app.api.core.models.snapshot import Snapshot as SnapshotModel
from qualibrate_app.api.core.schemas.state_updates import (
    StateUpdateRequestItems,
)
from qualibrate_app.api.core.types import DocumentSequenceType, IdType
from qualibrate_app.api.core.utils.request_utils import get_runner_config
from qualibrate_app.api.core.utils.types_parsing import types_conversion
from qualibrate_app.api.dependencies.search import get_search_path
from qualibrate_app.config import (
    get_settings,
)

snapshot_router = APIRouter(prefix="/snapshot/{id}", tags=["snapshot"])


def is_float(string: str) -> bool:
    try:
        float(string)
        return True
    except ValueError:
        return False


def _get_snapshot_instance(
    id: Annotated[IdType, Path()],
    settings: Annotated[QualibrateConfig, Depends(get_settings)],
) -> SnapshotBase:
    snapshot_types: dict[StorageType, type[SnapshotBase]] = {
        StorageType.local_storage: SnapshotLocalStorage,
        StorageType.timeline_db: SnapshotTimelineDb,
    }
    return snapshot_types[settings.storage.type](id=id, settings=settings)


@snapshot_router.get("/")
def get(
    *,
    load_type: SnapshotLoadType = SnapshotLoadType.Full,
    snapshot: Annotated[SnapshotBase, Depends(_get_snapshot_instance)],
) -> SnapshotModel:
    snapshot.load(load_type)
    return snapshot.dump()


@snapshot_router.get("/history")
def get_history(
    *,
    page: int = Query(1, gt=0),
    per_page: int = Query(50, gt=0),
    reverse: bool = False,
    global_reverse: bool = False,
    snapshot: Annotated[SnapshotBase, Depends(_get_snapshot_instance)],
) -> PagedCollection[SimplifiedSnapshotWithMetadata]:
    total, history = snapshot.get_latest_snapshots(
        page, per_page, global_reverse
    )
    history_dumped = [
        SimplifiedSnapshotWithMetadata(**snapshot.dump().model_dump())
        for snapshot in history
    ]
    if reverse:
        # TODO: make more correct relationship update
        history_dumped = list(reversed(history_dumped))
    return PagedCollection[SimplifiedSnapshotWithMetadata](
        page=page, per_page=per_page, total_items=total, items=history_dumped
    )


@snapshot_router.get("/compare")
def compare_by_id(
    id_to_compare: IdType,
    snapshot: Annotated[SnapshotBase, Depends(_get_snapshot_instance)],
) -> Mapping[str, Mapping[str, Any]]:
    return snapshot.compare_by_id(id_to_compare)


@snapshot_router.post("/update_entry")
def update_entry(
    *,
    snapshot: Annotated[SnapshotBase, Depends(_get_snapshot_instance)],
    data_path: Annotated[
        str,
        Body(
            ...,
            min_length=3,
            pattern="^#/.*",
            examples=["#/qubits/q0/frequency"],
        ),
    ],
    value: Annotated[Any, Body()],
    settings: Annotated[QualibrateConfig, Depends(get_settings)],
    qualibrate_token: Annotated[
        Union[str, None], Cookie(alias="Qualibrate-Token")
    ] = None,
) -> bool:
    cookies = (
        {"Qualibrate-Token": qualibrate_token}
        if qualibrate_token is not None
        else {}
    )
    type_ = snapshot.extract_state_update_type(data_path, cookies=cookies)
    if type_ is not None:
        value = types_conversion(value, type_)
    updated = snapshot.update_entry({data_path: value})
    if updated:
        try:
            runner_config = get_runner_config(settings)
        except RuntimeError as ex:
            logger.exception(str(ex))
            return False
        with contextlib.suppress(requests.exceptions.ConnectionError):
            requests.post(
                urljoin(runner_config.address_with_root, "record_state_update"),
                params={"key": data_path},
                cookies=cookies,
                timeout=runner_config.timeout,
            )
    return updated


@snapshot_router.post("/update_entries")
def update_entries(
    *,
    snapshot: Annotated[SnapshotBase, Depends(_get_snapshot_instance)],
    state_updates: StateUpdateRequestItems,
    settings: Annotated[QualibrateConfig, Depends(get_settings)],
    qualibrate_token: Annotated[
        Union[str, None], Cookie(alias="Qualibrate-Token")
    ] = None,
) -> bool:
    """
    Updates entries in a snapshot based on provided state updates.

    This endpoint extracts state update types for the provided paths and
    converts the update values according to the extracted types. It then applies
    the updates to the snapshot. If the update is successful, it attempts to
    record the state updates with the runner.
    """
    cookies = (
        {"Qualibrate-Token": qualibrate_token}
        if qualibrate_token is not None
        else {}
    )
    types = snapshot.extract_state_update_types(
        [item.data_path for item in state_updates.items], cookies=cookies
    )
    values = {
        item.data_path: types_conversion(
            item.value, cast(Mapping[str, Any], types[item.data_path])
        )
        for item in state_updates.items
        if types.get(item.data_path) is not None
    }
    updated = snapshot.update_entry(values)
    if updated:
        try:
            runner_config = get_runner_config(settings)
        except RuntimeError as ex:
            logger.exception(str(ex))
            return False
        for data_path in values:
            with contextlib.suppress(requests.exceptions.ConnectionError):
                requests.post(
                    urljoin(
                        runner_config.address_with_root, "record_state_update"
                    ),
                    params={"key": data_path},
                    cookies=cookies,
                    timeout=runner_config.timeout,
                )
    return updated


@snapshot_router.get("/search/data/values")
def search(
    snapshot: Annotated[SnapshotBase, Depends(_get_snapshot_instance)],
    data_path: Annotated[list[Union[str, int]], Depends(get_search_path)],
) -> Optional[DocumentSequenceType]:
    return snapshot.search(data_path, load=True)


@snapshot_router.get("/search/data/value/any_depth")
def search_recursive(
    snapshot: Annotated[SnapshotBase, Depends(_get_snapshot_instance)],
    target_key: str,
) -> Optional[DocumentSequenceType]:
    return snapshot.search_recursive(target_key, load=True)



from typing import Annotated

from fastapi import APIRouter, Depends, Path, Query
from qualibrate_config.models import QualibrateConfig, StorageType

from qualibrate_app.api.core.domain.bases.branch import (
    BranchBase,
    BranchLoadType,
)
from qualibrate_app.api.core.domain.bases.node import NodeLoadType
from qualibrate_app.api.core.domain.bases.snapshot import SnapshotLoadType
from qualibrate_app.api.core.domain.local_storage.branch import (
    BranchLocalStorage,
)
from qualibrate_app.api.core.domain.timeline_db.branch import BranchTimelineDb
from qualibrate_app.api.core.models.branch import Branch as BranchModel
from qualibrate_app.api.core.models.node import Node as NodeModel
from qualibrate_app.api.core.models.paged import PagedCollection
from qualibrate_app.api.core.models.snapshot import (
    SimplifiedSnapshotWithMetadata,
)
from qualibrate_app.api.core.models.snapshot import Snapshot as SnapshotModel
from qualibrate_app.api.core.types import IdType
from qualibrate_app.config import get_settings

branch_router = APIRouter(prefix="/branch/{name}", tags=["branch"])


def _get_branch_instance(
    name: Annotated[str, Path()],
    settings: Annotated[QualibrateConfig, Depends(get_settings)],
) -> BranchBase:
    branch_types = {
        StorageType.local_storage: BranchLocalStorage,
        StorageType.timeline_db: BranchTimelineDb,
    }
    return branch_types[settings.storage.type](name=name, settings=settings)


@branch_router.get("/")
def get(
    *,
    load_type: BranchLoadType = BranchLoadType.Full,
    branch: Annotated[BranchBase, Depends(_get_branch_instance)],
) -> BranchModel:
    branch.load(load_type)
    return branch.dump()


@branch_router.get("/snapshot")
def get_snapshot(
    *,
    snapshot_id: IdType,
    load_type: SnapshotLoadType = SnapshotLoadType.Metadata,
    branch: Annotated[BranchBase, Depends(_get_branch_instance)],
) -> SnapshotModel:
    snapshot = branch.get_snapshot(snapshot_id)
    snapshot.load(load_type)
    return snapshot.dump()


@branch_router.get("/snapshot/latest")
def get_latest_snapshot(
    *,
    load_type: SnapshotLoadType = SnapshotLoadType.Metadata,
    branch: Annotated[BranchBase, Depends(_get_branch_instance)],
) -> SnapshotModel:
    snapshot = branch.get_snapshot()
    snapshot.load(load_type)
    return snapshot.dump()


@branch_router.get("/node")
def get_node(
    *,
    node_id: int,
    load_type: NodeLoadType = NodeLoadType.Full,
    branch: Annotated[BranchBase, Depends(_get_branch_instance)],
) -> NodeModel:
    node = branch.get_node(node_id)
    node.load(load_type)
    return node.dump()


@branch_router.get("/node/latest")
def get_latest_node(
    *,
    load_type: NodeLoadType = NodeLoadType.Full,
    branch: Annotated[BranchBase, Depends(_get_branch_instance)],
) -> NodeModel:
    node = branch.get_node()
    node.load(load_type)
    return node.dump()


@branch_router.get("/snapshots_history")
def get_snapshots_history(
    *,
    page: int = Query(1, gt=0),
    per_page: int = Query(50, gt=0),
    reverse: bool = False,
    global_reverse: bool = False,
    branch: Annotated[BranchBase, Depends(_get_branch_instance)],
) -> PagedCollection[SimplifiedSnapshotWithMetadata]:
    total, snapshots = branch.get_latest_snapshots(
        page, per_page, global_reverse
    )
    snapshots_dumped = [
        SimplifiedSnapshotWithMetadata(**snapshot.dump().model_dump())
        for snapshot in snapshots
    ]
    if reverse:
        # TODO: make more correct relationship update
        snapshots_dumped = list(reversed(snapshots_dumped))
    return PagedCollection[SimplifiedSnapshotWithMetadata](
        page=page,
        per_page=per_page,
        total_items=total,
        items=snapshots_dumped,
    )


@branch_router.get("/nodes_history")
def get_nodes_history(
    *,
    page: int = Query(1, gt=0),
    per_page: int = Query(50, gt=0),
    reverse: bool = False,
    global_reverse: bool = False,
    branch: Annotated[BranchBase, Depends(_get_branch_instance)],
) -> PagedCollection[NodeModel]:
    total, nodes = branch.get_latest_nodes(page, per_page, global_reverse)
    nodes_dumped = [node.dump() for node in nodes]
    if reverse:
        # TODO: make more correct relationship update
        nodes_dumped = list(reversed(nodes_dumped))
    return PagedCollection[NodeModel](
        page=page,
        per_page=per_page,
        total_items=total,
        items=nodes_dumped,
    )



import string
from collections.abc import Mapping, Sequence
from typing import Any

AllowedSearchKeys = string.ascii_letters + string.digits + "-_*"

IdType = int
DocumentType = Mapping[str, Any]
DocumentSequenceType = Sequence[DocumentType]



from collections.abc import Generator
from datetime import date
from functools import partial
from pathlib import Path
from typing import Any, Callable, Optional

from qualibrate_app.api.core.domain.local_storage._id_to_local_path import (
    IdToLocalPath,
)
from qualibrate_app.api.core.types import IdType
from qualibrate_app.api.core.utils.path.node import NodePath
from qualibrate_app.api.core.utils.path.node_date import NodesDatePath


def find_latest_node(base_path: Path) -> Optional[NodePath]:
    return max(
        map(NodePath, base_path.glob("*/#*")),
        key=lambda p: p.id or -1,  # type: ignore[union-attr]
        default=None,
    )


def find_latest_node_id(base_path: Path) -> IdType:
    latest_node = find_latest_node(base_path)
    return (latest_node.id or -1) if latest_node else -1


def _validate_date_range(
    date_path: NodesDatePath,
    min_date: Optional[date] = None,
    max_date: Optional[date] = None,
) -> bool:
    try:
        dt = date_path.date
        if min_date is None and max_date is None:
            return True
        if min_date is not None and max_date is not None:
            return min_date <= dt <= max_date
        if min_date is not None:
            return min_date <= dt
        if max_date is not None:
            return dt <= max_date
        raise ValueError()
    except ValueError:
        return False


def _validate_node_id(
    node_path: NodePath, min_id: IdType, max_id: IdType
) -> bool:
    id_ = node_path.id
    return id_ is not None and (min_id <= id_ <= max_id)


def find_n_latest_nodes_ids(
    base_path: Path,
    page: int,
    per_page: int,
    project_name: str,
    max_node_id: Optional[int] = None,
) -> Generator[IdType, None, None]:
    """
    Generator of n latest nodes ids
    """
    n = per_page
    page -= 1

    max_node_id = (
        max_node_id
        if max_node_id is not None
        else find_latest_node_id(base_path)
    )
    node_id_max_val = max(0, max_node_id - page * per_page)
    if node_id_max_val == 0:
        return None
    node_id_min_val = max(1, node_id_max_val - per_page + 1)

    next_ = None
    paths_mapping = IdToLocalPath()
    get_node_path = partial(
        paths_mapping.get,
        project=project_name,
        project_path=base_path,
    )
    min_node_path = get_node_path(id=node_id_min_val)
    min_node_path_date = (
        min_node_path.date if min_node_path is not None else None
    )
    max_node_path = get_node_path(id=node_id_max_val)
    max_node_path_date = (
        max_node_path.date if max_node_path is not None else None
    )
    date_filters: list[
        tuple[Callable[[NodesDatePath], bool], tuple[Any, ...]]
    ] = [
        (Path.is_dir, tuple()),
        (_validate_date_range, (min_node_path_date, max_node_path_date)),
    ]
    node_filters = [(_validate_node_id, (node_id_min_val, node_id_max_val))]
    for node_date in sorted(
        filter(
            lambda p: all(filter_(p, *args) for filter_, args in date_filters),
            map(NodesDatePath, base_path.iterdir()),
        ),
        reverse=True,
    ):
        node_paths = filter(
            lambda p: all(filter_(p, *args) for filter_, args in node_filters),
            map(NodePath, node_date.glob("#*")),
        )
        node_path_ids = {
            int(path.stem[1:].split("_")[0]): path for path in node_paths
        }
        for _, node in sorted(
            node_path_ids.items(),
            key=lambda x: x[0],
            reverse=True,
        ):
            current = next_
            if current is None:
                next_ = node.id
                continue
            next_ = node.id
            n -= 1
            if n > 0:
                yield current
            else:
                # we have found all needed nodes; just exit from generator
                yield current
                return None
    # all nodes are checked;
    # if last item has already been yielded then just exits
    if next_ is None:
        return None
    # if one more item still exists then yielded it and then exists
    yield next_
    return None



from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from datetime import datetime
from enum import IntEnum
from typing import Any, ClassVar, Optional, Union

from qualibrate_config.models import QualibrateConfig

from qualibrate_app.api.core.domain.bases.base_with_settings import (
    DomainWithConfigBase,
)
from qualibrate_app.api.core.domain.bases.i_dump import IDump
from qualibrate_app.api.core.models.snapshot import Snapshot as SnapshotModel
from qualibrate_app.api.core.types import (
    DocumentSequenceType,
    DocumentType,
    IdType,
)
from qualibrate_app.api.core.utils.find_utils import (
    get_subpath_value_on_any_depth,
)

__all__ = ["SnapshotBase", "SnapshotLoadType"]


class SnapshotLoadType(IntEnum):
    Empty = 0
    Minified = 1
    Metadata = 2
    Data = 3
    Full = 4


class SnapshotBase(DomainWithConfigBase, IDump, ABC):
    _items_keys: ClassVar[tuple[str, ...]] = ("data", "metadata")

    def __init__(
        self,
        id: IdType,
        content: Optional[DocumentType] = None,
        *,
        settings: QualibrateConfig,
    ):
        super().__init__(settings)
        self._id = id
        if content is None:
            self._load_type = SnapshotLoadType.Empty
            self.content = {}
            return
        specified_items_keys = {
            key: key in content for key in self.__class__._items_keys
        }
        if any(specified_items_keys.values()):
            if all(specified_items_keys.values()):
                self._load_type = SnapshotLoadType.Full
            elif specified_items_keys["data"]:
                self._load_type = SnapshotLoadType.Data
            else:
                self._load_type = SnapshotLoadType.Metadata
        else:
            self._load_type = SnapshotLoadType.Minified
        self.content = dict(content)

    @abstractmethod
    def load(self, load_type: SnapshotLoadType) -> None:
        pass

    @property
    def load_type(self) -> SnapshotLoadType:
        return self._load_type

    @property
    def id(self) -> Optional[IdType]:
        return self._id

    @property
    @abstractmethod
    def created_at(self) -> Optional[datetime]:
        pass

    @property
    @abstractmethod
    def parents(self) -> Optional[list[IdType]]:
        pass

    @property
    def metadata(self) -> Optional[DocumentType]:
        return self.content.get("metadata")

    @property
    def data(self) -> Optional[DocumentType]:
        return self.content.get("data")

    @abstractmethod
    def search(
        self,
        search_path: Sequence[Union[str, int]],
        load: bool = False,
    ) -> Optional[DocumentSequenceType]:
        pass

    def search_recursive(
        self, target_key: str, load: bool = False
    ) -> Optional[DocumentSequenceType]:
        if self._load_type < SnapshotLoadType.Data and not load:
            return None
        self.load(SnapshotLoadType.Data)
        # TODO: update logic; not use quam
        data = (self.data or {}).get("quam")
        if data is None:
            return None
        return get_subpath_value_on_any_depth(data, target_key)

    @abstractmethod
    def get_latest_snapshots(
        self, page: int = 1, per_page: int = 50, reverse: bool = False
    ) -> tuple[int, Sequence["SnapshotBase"]]:
        pass

    @abstractmethod
    def compare_by_id(
        self, other_snapshot_int: int
    ) -> Mapping[str, Mapping[str, Any]]:
        pass

    def dump(self) -> SnapshotModel:
        return SnapshotModel(**self.content)

    @abstractmethod
    def extract_state_update_type(
        self,
        path: str,
        **kwargs: Mapping[str, Any],
    ) -> Optional[Mapping[str, Any]]:
        pass

    @abstractmethod
    def extract_state_update_types(
        self,
        paths: Sequence[str],
        **kwargs: Mapping[str, Any],
    ) -> Mapping[str, Optional[Mapping[str, Any]]]:
        pass

    @abstractmethod
    def update_entry(self, updates: Mapping[str, Any]) -> bool:
        pass



import math
from typing import Generic, TypeVar

from pydantic import BaseModel, computed_field

ItemT = TypeVar("ItemT")


class PagedCollection(BaseModel, Generic[ItemT]):
    page: int
    per_page: int
    total_items: int
    items: list[ItemT]

    @computed_field
    def total_pages(self) -> int:
        return math.ceil(self.total_items / self.per_page)



from qualibrate_app.api.core.utils.path.node import NodePath
from qualibrate_app.api.core.utils.path.node_date import NodesDatePath


def date_less_or_eq(date_path: NodesDatePath, date_to_compare: str) -> bool:
    # TODO: use date to compare
    return date_path.stem <= date_to_compare


def id_less_then_snapshot(node_path: NodePath, node_id_to_compare: int) -> bool:
    node_id = node_path.id
    if node_id is None:
        return False
    return int(node_id) < node_id_to_compare



from collections.abc import Sequence
from datetime import datetime
from pathlib import Path

from qualibrate_app.api.core.domain.bases.project import ProjectsManagerBase
from qualibrate_app.api.core.models.project import Project
from qualibrate_app.api.core.utils.path.node import NodePath
from qualibrate_app.api.exceptions.classes.storage import QFileNotFoundException
from qualibrate_app.api.exceptions.classes.values import QValueException


class ProjectsManagerLocalStorage(ProjectsManagerBase):
    def _active_project_setter(self, value: str) -> None:
        self._set_user_storage_project(value)

    def _set_user_storage_project(self, project_name: str) -> None:
        new_project_path = self._resolve_new_project_path(
            project_name,
            self._settings.project,
            self._settings.storage.location,
        )
        if not new_project_path.is_dir():
            raise QFileNotFoundException(
                f"Project {project_name} does not exist"
            )
        super()._set_user_storage_project(project_name)

    def create(self, project_name: str) -> str:
        new_project_path = self._resolve_new_project_path(
            project_name,
            self._settings.project,
            self._settings.storage.location,
        )
        if new_project_path.is_dir():
            raise QValueException(f"Project {project_name} already exists.")
        new_project_path.mkdir(parents=True)
        return project_name

    def _get_project_info(self, project_path: Path) -> Project:
        project_created_at = datetime.fromtimestamp(
            project_path.stat().st_mtime
        ).astimezone()
        max_node_number_path = max(
            map(NodePath, project_path.glob("*/#*")),
            key=lambda n: n.id or -1,
            default=None,
        )
        if max_node_number_path is not None:
            (
                max_node_num_id,
                _,
                max_node_number_time,
            ) = max_node_number_path.get_node_id_name_time()
        else:
            max_node_num_id = None
            max_node_number_time = None
        max_node_number = max_node_num_id or -1
        if (
            max_node_number_path is not None
            and max_node_number_time is not None
        ):
            last_modified_ts = max_node_number_path.datetime.replace(
                hour=max_node_number_time.hour,
                minute=max_node_number_time.minute,
                second=max_node_number_time.second,
            ).astimezone()
        else:
            last_modified_node_path = max(
                project_path.glob("*/#*"),
                key=lambda n: n.stat().st_mtime,
                default=None,
            )
            last_modified_ts = (
                datetime.fromtimestamp(
                    last_modified_node_path.stat().st_mtime
                ).astimezone()
                if last_modified_node_path is not None
                else project_created_at
            )
        return Project(
            name=project_path.name,
            nodes_number=max_node_number,
            created_at=project_created_at,
            last_modified_at=last_modified_ts,
        )

    def list(self) -> Sequence[Project]:
        base_path = self._resolve_base_projects_path(
            self._settings.project,
            self._settings.storage.location,
        )
        return [
            self._get_project_info(p)
            for p in filter(Path.is_dir, base_path.iterdir())
        ]



from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from copy import deepcopy
from pathlib import Path
from typing import Any

from qualibrate_config.file import read_config_file
from qualibrate_config.models import QualibrateConfig
from qualibrate_config.references.resolvers import resolve_references
from qualibrate_config.vars import QUALIBRATE_CONFIG_KEY

from qualibrate_app.api.core.domain.bases.base_with_settings import (
    DomainWithConfigBase,
)
from qualibrate_app.api.core.models.project import Project


class ProjectsManagerBase(DomainWithConfigBase, ABC):
    def __init__(self, settings: QualibrateConfig, config_path: Path):
        super().__init__(settings)
        self._config_path = config_path

    @property
    def project(self) -> str:
        return self._settings.project

    @project.setter
    def project(self, value: str) -> None:
        self._active_project_setter(value)

    @abstractmethod
    def _active_project_setter(self, value: str) -> None:
        pass

    def _set_user_storage_project(self, project_name: str) -> None:
        self._settings.project = project_name

    @abstractmethod
    def create(self, project_name: str) -> str:
        pass

    @abstractmethod
    def list(self) -> Sequence[Project]:
        pass

    def _get_raw_and_resolved_ref_config(
        self, project_name: str
    ) -> tuple[Mapping[str, Any], Mapping[str, Any]]:
        raw_config = read_config_file(self._config_path, solve_references=False)
        # TODO: over way to update project;
        old_project_name = raw_config[QUALIBRATE_CONFIG_KEY]["project"]
        if old_project_name == project_name:
            return raw_config, deepcopy(raw_config)
        raw_config[QUALIBRATE_CONFIG_KEY]["project"] = project_name
        new_config = resolve_references(raw_config)
        if (
            new_config[QUALIBRATE_CONFIG_KEY]["storage"]["location"]
            == raw_config[QUALIBRATE_CONFIG_KEY]["storage"]["location"]
        ):
            # there is no reference in config
            project_path = Path(
                new_config[QUALIBRATE_CONFIG_KEY]["storage"]["location"]
            )
            new_project_path = self._resolve_new_project_path(
                project_name, old_project_name, project_path
            )
            new_config[QUALIBRATE_CONFIG_KEY]["storage"]["location"] = str(
                new_project_path
            )
        return raw_config, new_config

    @classmethod
    def _resolve_base_projects_path(
        cls, project_name: str, user_storage: Path
    ) -> Path:
        if project_name not in user_storage.parts:
            # project name isn't part of user storage path;
            # use parent dir as base path.
            return user_storage.parent
        project_name_index_from_end = tuple(reversed(user_storage.parts)).index(
            project_name
        )
        return user_storage.parents[project_name_index_from_end]

    @classmethod
    def _resolve_new_project_path(
        cls,
        new_project_name: str,
        current_project_name: str,
        current_user_storage: Path,
    ) -> Path:
        if current_project_name not in current_user_storage.parts:
            # project name isn't part of user storage path;
            # use neighbour dir as new project path.
            return current_user_storage.parent / new_project_name
        project_name_index_from_start = current_user_storage.parts.index(
            current_project_name
        )
        base_project_path = cls._resolve_base_projects_path(
            current_project_name, current_user_storage
        )
        return base_project_path.joinpath(
            new_project_name,
            *current_user_storage.parts[project_name_index_from_start + 1 :],
        )



from collections.abc import Sequence
from typing import Annotated, Any, Union

from fastapi import APIRouter, Depends, Query
from qualibrate_config.models import QualibrateConfig, StorageType

from qualibrate_app.api.core.domain.bases.branch import BranchLoadType
from qualibrate_app.api.core.domain.bases.node import NodeLoadType
from qualibrate_app.api.core.domain.bases.root import RootBase
from qualibrate_app.api.core.domain.bases.snapshot import SnapshotLoadType
from qualibrate_app.api.core.domain.local_storage.root import RootLocalStorage
from qualibrate_app.api.core.domain.timeline_db.root import RootTimelineDb
from qualibrate_app.api.core.models.branch import Branch as BranchModel
from qualibrate_app.api.core.models.node import Node as NodeModel
from qualibrate_app.api.core.models.paged import PagedCollection
from qualibrate_app.api.core.models.snapshot import (
    SimplifiedSnapshotWithMetadata,
)
from qualibrate_app.api.core.models.snapshot import Snapshot as SnapshotModel
from qualibrate_app.api.core.types import IdType
from qualibrate_app.api.dependencies.search import get_search_path
from qualibrate_app.config import (
    get_settings,
)

root_router = APIRouter(prefix="/root", tags=["root"])


def _get_root_instance(
    settings: Annotated[QualibrateConfig, Depends(get_settings)],
) -> RootBase:
    root_types = {
        StorageType.local_storage: RootLocalStorage,
        StorageType.timeline_db: RootTimelineDb,
    }
    return root_types[settings.storage.type](settings=settings)


@root_router.get("/branch")
def get_branch(
    *,
    branch_name: str = "main",
    load_type: BranchLoadType = BranchLoadType.Full,
    root: Annotated[RootBase, Depends(_get_root_instance)],
) -> BranchModel:
    branch = root.get_branch(branch_name)
    branch.load(load_type)
    return branch.dump()


@root_router.get("/node")
def get_node_by_id(
    *,
    id: IdType,
    load_type: NodeLoadType = NodeLoadType.Full,
    root: Annotated[RootBase, Depends(_get_root_instance)],
) -> NodeModel:
    node = root.get_node(id)
    node.load(load_type)
    return node.dump()


@root_router.get("/node/latest")
def get_latest_node(
    *,
    load_type: NodeLoadType = NodeLoadType.Full,
    root: Annotated[RootBase, Depends(_get_root_instance)],
) -> NodeModel:
    node = root.get_node()
    node.load(load_type)
    return node.dump()


@root_router.get("/snapshot")
def get_snapshot_by_id(
    *,
    id: IdType,
    load_type: SnapshotLoadType = SnapshotLoadType.Metadata,
    root: Annotated[RootBase, Depends(_get_root_instance)],
) -> SnapshotModel:
    snapshot = root.get_snapshot(id)
    snapshot.load(load_type)
    return snapshot.dump()


@root_router.get("/snapshot/latest")
def get_latest_snapshot(
    *,
    load_type: SnapshotLoadType = SnapshotLoadType.Metadata,
    root: Annotated[RootBase, Depends(_get_root_instance)],
) -> SnapshotModel:
    snapshot = root.get_snapshot()
    snapshot.load(load_type)
    return snapshot.dump()


@root_router.get("/snapshots_history")
def get_snapshots_history(
    *,
    page: int = Query(1, gt=0),
    per_page: int = Query(50, gt=0),
    reverse: bool = False,
    global_reverse: bool = False,
    root: Annotated[RootBase, Depends(_get_root_instance)],
) -> PagedCollection[SimplifiedSnapshotWithMetadata]:
    total, snapshots = root.get_latest_snapshots(page, per_page, global_reverse)
    snapshots_dumped = [
        SimplifiedSnapshotWithMetadata(**snapshot.dump().model_dump())
        for snapshot in snapshots
    ]
    if reverse:
        snapshots_dumped = list(reversed(snapshots_dumped))
    return PagedCollection[SimplifiedSnapshotWithMetadata](
        page=page,
        per_page=per_page,
        total_items=total,
        items=snapshots_dumped,
    )


@root_router.get("/nodes_history")
def get_nodes_history(
    *,
    page: int = Query(1, gt=0),
    per_page: int = Query(50, gt=0),
    reverse: bool = False,
    global_reverse: bool = False,
    root: Annotated[RootBase, Depends(_get_root_instance)],
) -> PagedCollection[NodeModel]:
    total, nodes = root.get_latest_nodes(page, per_page, global_reverse)
    nodes_dumped = [node.dump() for node in nodes]
    if reverse:
        # TODO: make more correct relationship update
        nodes_dumped = list(reversed(nodes_dumped))
    return PagedCollection[NodeModel](
        page=page,
        per_page=per_page,
        total_items=total,
        items=nodes_dumped,
    )


@root_router.get("/search")
def search_snapshot(
    id: IdType,
    data_path: Annotated[Sequence[Union[str, int]], Depends(get_search_path)],
    root: Annotated[RootBase, Depends(_get_root_instance)],
) -> Any:
    return root.search_snapshot(id, data_path)



from collections.abc import Mapping, Sequence
from itertools import chain
from typing import Any, Callable, Optional, Union, cast

from qualibrate_app.api.core.types import DocumentSequenceType


def _get_subpath_value_wildcard(
    obj: Union[Mapping[str, Any], Sequence[Any]],
    target_path: Sequence[Union[str, int]],
    current_path: list[Union[str, int]],
) -> DocumentSequenceType:
    if len(target_path) == 1:
        if isinstance(obj, Sequence):
            return [
                {"key": current_path + [i], "value": val}
                for i, val in enumerate(obj)
            ]
        elif isinstance(obj, Mapping):
            return [
                {"key": current_path + [key], "value": val}
                for key, val in obj.items()
            ]
        else:
            return []

    if not isinstance(obj, (Mapping, Sequence)):
        return []
    iter_function = cast(
        Callable[
            ..., Union[Sequence[tuple[str, Any]], Sequence[tuple[int, Any]]]
        ],
        enumerate if isinstance(obj, Sequence) else dict.items,
    )
    return list(
        chain.from_iterable(
            get_subpath_value(value, target_path[1:], current_path + [idx])
            for idx, value in iter_function(obj)
        )
    )


def _check_key_valid(obj: Any, key: Union[str, int]) -> bool:
    if isinstance(obj, Sequence):
        return isinstance(key, int) and 0 <= key < len(obj)
    if isinstance(obj, Mapping):
        return isinstance(key, str) and key in obj
    return False


def get_subpath_value_mapping(
    obj: Mapping[str, Any],
    target_path: Sequence[Union[str, int]],
    current_path: list[Union[str, int]],
    key: str,
) -> DocumentSequenceType:
    if len(target_path) == 1:
        return [{"key": current_path + [key], "value": obj[key]}]
    else:
        return get_subpath_value(
            obj[key], target_path[1:], current_path + [key]
        )


def get_subpath_value_sequence(
    obj: Sequence[Any],
    target_path: Sequence[Union[str, int]],
    current_path: list[Union[str, int]],
    key: int,
) -> DocumentSequenceType:
    if len(target_path) == 1:
        return [{"key": current_path + [key], "value": obj[key]}]
    else:
        return get_subpath_value(
            obj[key], target_path[1:], current_path + [key]
        )


def get_subpath_value(
    obj: Union[Mapping[str, Any], Sequence[Any]],
    target_path: Sequence[Union[str, int]],
    current_path: Optional[list[Union[str, int]]] = None,
) -> DocumentSequenceType:
    if current_path is None:
        current_path = []
    if len(target_path) == 0:
        return []
    key = target_path[0]
    if key == "*":
        return _get_subpath_value_wildcard(obj, target_path, current_path)
    key_valid = _check_key_valid(obj, key)
    if not key_valid:
        return []
    if isinstance(obj, Mapping):
        return get_subpath_value_mapping(
            obj, target_path, current_path, cast(str, key)
        )
    return get_subpath_value_sequence(
        obj, target_path, current_path, cast(int, key)
    )


def get_subpath_value_on_any_depth(
    obj: Union[Mapping[str, Any], Sequence[Any]],
    key: str,
    current_path: Optional[list[Union[str, int]]] = None,
    paths: Optional[list[dict[str, Any]]] = None,
) -> list[dict[str, Any]]:
    if current_path is None:
        current_path = []
    if paths is None:
        paths = []
    if isinstance(obj, Mapping):
        for k, v in obj.items():
            if k == key:
                paths.append({"path": current_path + [k], "value": v})
            else:
                get_subpath_value_on_any_depth(
                    v, key, current_path + [k], paths
                )
    elif isinstance(obj, Sequence) and not isinstance(obj, str):
        for i, item in enumerate(obj):
            get_subpath_value_on_any_depth(item, key, current_path + [i], paths)
    return paths



from fastapi import APIRouter

from qualibrate_app.api.routes.branch import branch_router
from qualibrate_app.api.routes.data_file import data_file_router
from qualibrate_app.api.routes.project import project_router
from qualibrate_app.api.routes.root import root_router
from qualibrate_app.api.routes.snapshot import snapshot_router

__all__ = ["project_router", "storage_router"]

storage_router = APIRouter()

storage_router.include_router(root_router)
storage_router.include_router(branch_router)
storage_router.include_router(data_file_router)
storage_router.include_router(snapshot_router)



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any, Optional, Union

from qualibrate_config.models import QualibrateConfig

from qualibrate_app.api.core.domain.bases.base_with_settings import (
    DomainWithConfigBase,
)
from qualibrate_app.api.core.domain.bases.branch import BranchBase
from qualibrate_app.api.core.domain.bases.node import NodeBase
from qualibrate_app.api.core.domain.bases.snapshot import SnapshotBase
from qualibrate_app.api.core.types import IdType

__all__ = ["RootBase"]


class RootBase(DomainWithConfigBase, ABC):
    def __init__(self, settings: QualibrateConfig):
        super().__init__(settings)

    @abstractmethod
    def get_branch(self, branch_name: str) -> BranchBase:
        pass

    @abstractmethod
    def get_snapshot(self, id: Optional[IdType] = None) -> SnapshotBase:
        pass

    @abstractmethod
    def get_node(self, id: Optional[IdType] = None) -> NodeBase:
        pass

    @abstractmethod
    def get_latest_snapshots(
        self,
        page: int = 1,
        per_page: int = 50,
        reverse: bool = False,
    ) -> tuple[int, Sequence[SnapshotBase]]:
        pass

    @abstractmethod
    def get_latest_nodes(
        self,
        page: int = 1,
        per_page: int = 50,
        reverse: bool = False,
    ) -> tuple[int, Sequence[NodeBase]]:
        pass

    @abstractmethod
    def search_snapshot(
        self,
        snapshot_id: IdType,
        data_path: Sequence[Union[str, int]],
    ) -> Any:
        pass



from fastapi import APIRouter

from qualibrate_app.api.routes import project_router, storage_router

api_router = APIRouter()


@api_router.get("/")
def ping() -> str:
    return "pong"


api_router.include_router(project_router)
api_router.include_router(storage_router)



from collections.abc import Mapping
from pathlib import Path
from typing import Any, Optional

from pydantic import BaseModel


class Storage(BaseModel):
    path: Optional[Path] = None
    data: Optional[Mapping[str, Any]] = None



from pydantic import AwareDatetime, BaseModel, field_serializer

from qualibrate_app.api.core.types import IdType

__all__ = ["ModelWithId", "ModelCreatedAt", "ModelWithIdCreatedAt"]


class ModelWithId(BaseModel):
    id: IdType


class ModelCreatedAt(BaseModel):
    created_at: AwareDatetime

    @field_serializer("created_at")
    def dt_serializer(self, dt: AwareDatetime) -> str:
        return dt.isoformat(timespec="seconds")


class ModelWithIdCreatedAt(ModelWithId, ModelCreatedAt):
    pass



from collections.abc import Mapping, Sequence
from datetime import datetime
from typing import Any, Optional, Union, cast

from qualibrate_config.models import (
    QualibrateConfig,
)

from qualibrate_app.api.core.domain.bases.snapshot import (
    SnapshotBase,
    SnapshotLoadType,
)
from qualibrate_app.api.core.types import (
    DocumentSequenceType,
    DocumentType,
    IdType,
)
from qualibrate_app.api.core.utils.find_utils import get_subpath_value
from qualibrate_app.api.core.utils.request_utils import request_with_db
from qualibrate_app.api.core.utils.snapshots_compare import jsonpatch_to_mapping
from qualibrate_app.api.exceptions.classes.timeline_db import QJsonDbException

__all__ = ["SnapshotTimelineDb"]


class SnapshotTimelineDb(SnapshotBase):
    def __init__(
        self,
        id: IdType,
        content: Optional[DocumentType] = None,
        *,
        settings: QualibrateConfig,
    ):
        super().__init__(id, content, settings=settings)

    def load(self, load_type: SnapshotLoadType) -> None:
        if load_type <= self._load_type:
            return None
        fields: Optional[list[str]] = ["id", "_id", "parents", "created_at"]
        if fields is not None and load_type == SnapshotLoadType.Metadata:
            fields.append("metadata")
        elif load_type >= SnapshotLoadType.Data:
            fields = None
        params = None if fields is None else {"fields": fields}
        timeline_db_config = self.timeline_db_config
        result = request_with_db(
            f"snapshot/{self.id}/",
            params=params,
            db_name=self._settings.project,
            host=timeline_db_config.address_with_root,
            timeout=timeline_db_config.timeout,
        )
        no_snapshot_ex = QJsonDbException("Snapshot data wasn't retrieved.")
        if result.status_code != 200:
            raise no_snapshot_ex
        content = result.json()
        if content is None:
            raise no_snapshot_ex
        if fields is None or "metadata" in fields:  # metadata was requested
            content["metadata"] = content.get("metadata", {})
            self._load_type = SnapshotLoadType.Metadata
        if fields is None:  # data was requested
            content["data"] = content.get("data", {})
            self._load_type = SnapshotLoadType.Full
        self.content.update(content)

    @property
    def load_type(self) -> SnapshotLoadType:
        return self._load_type

    @property
    def id(self) -> Optional[IdType]:
        return self._id

    @property
    def created_at(self) -> Optional[datetime]:
        if "created_at" not in self.content:
            return None
        return datetime.fromisoformat(str(self.content.get("created_at")))

    @property
    def parents(self) -> Optional[list[IdType]]:
        return self.content.get("parents")

    def search(
        self,
        search_path: Sequence[Union[str, int]],
        load: bool = False,
    ) -> Optional[DocumentSequenceType]:
        """Make search in current instance of Snapshot."""
        if self._load_type < SnapshotLoadType.Data and not load:
            return None
        self.load(SnapshotLoadType.Data)
        data = self.data
        if data is None:
            return None
        return get_subpath_value(data, search_path)

    def get_latest_snapshots(
        self, page: int = 1, per_page: int = 50, reverse: bool = False
    ) -> tuple[int, Sequence[SnapshotBase]]:
        timeline_db_config = self.timeline_db_config
        result = request_with_db(
            f"snapshot/{self.id}/history",
            params={"page": page, "per_page": per_page, "reverse": reverse},
            db_name=self._settings.project,
            host=timeline_db_config.address_with_root,
            timeout=timeline_db_config.timeout,
        )
        if result.status_code != 200:
            raise QJsonDbException("Snapshot history wasn't retrieved.")
        data = dict(result.json())

        return (
            cast(int, data["total"]),
            [
                SnapshotTimelineDb(
                    int(snapshot["id"]), snapshot, settings=self._settings
                )
                for snapshot in data["items"]
            ],
        )

    def compare_by_id(
        self, other_snapshot_int: int
    ) -> Mapping[str, Mapping[str, Any]]:
        if self.id == other_snapshot_int:
            return {}
        timeline_db_config = self.timeline_db_config
        response = request_with_db(
            "action/compare",
            params={"left_id": self.id, "right_id": other_snapshot_int},
            db_name=self._settings.project,
            host=timeline_db_config.address_with_root,
            timeout=timeline_db_config.timeout,
        )
        if response.status_code != 200:
            raise QJsonDbException("Difference wasn't retrieved.")
        result = dict(response.json())
        original = dict(result["original"])
        patch = result.get("patch")
        if patch is None:
            return {}
        return jsonpatch_to_mapping(
            original, cast(Sequence[Mapping[str, Any]], patch)
        )

    def update_entry(self, updates: Mapping[str, Any]) -> bool:
        # TODO: update timeline db snapshot entry
        return False

    def extract_state_update_type(
        self,
        path: str,
        **kwargs: Mapping[str, Any],
    ) -> Optional[Mapping[str, Any]]:
        return None

    def extract_state_update_types(
        self,
        paths: Sequence[str],
        **kwargs: Mapping[str, Any],
    ) -> Mapping[str, Optional[Mapping[str, Any]]]:
        return {}



from qualibrate_config.models import QualibrateConfig

from qualibrate_app.api.core.domain.bases.node import NodeBase
from qualibrate_app.api.core.domain.local_storage.snapshot import (
    SnapshotLocalStorage,
)
from qualibrate_app.api.core.types import IdType

__all__ = ["NodeLocalStorage"]


class NodeLocalStorage(NodeBase):
    def __init__(self, node_id: IdType, settings: QualibrateConfig):
        super().__init__(
            node_id, SnapshotLocalStorage(node_id, settings=settings), settings
        )



from typing import Annotated, Optional, Union

from fastapi import APIRouter, Depends, Path
from qualibrate_config.models import QualibrateConfig, StorageType

from qualibrate_app.api.core.domain.bases.node import NodeLoadType
from qualibrate_app.api.core.domain.bases.storage import (
    DataFileStorage,
    StorageLoadType,
)
from qualibrate_app.api.core.domain.local_storage.node import NodeLocalStorage
from qualibrate_app.api.core.domain.timeline_db.node import NodeTimelineDb
from qualibrate_app.api.core.types import DocumentType
from qualibrate_app.api.core.utils.request_utils import HTTPException422
from qualibrate_app.config import (
    get_settings,
)

data_file_router = APIRouter(prefix="/data_file/{node_id}", tags=["data file"])


def _get_storage_instance(
    node_id: Annotated[int, Path()],
    settings: Annotated[QualibrateConfig, Depends(get_settings)],
) -> DataFileStorage:
    node_types: dict[
        StorageType, Union[type[NodeLocalStorage], type[NodeTimelineDb]]
    ] = {
        StorageType.local_storage: NodeLocalStorage,
        StorageType.timeline_db: NodeTimelineDb,
    }
    node = node_types[settings.storage.type](node_id, settings=settings)
    try:
        node.load(NodeLoadType.Full)
    except NotADirectoryError as e:
        raise HTTPException422(detail=str(e)) from None
    if node.storage is None:
        raise HTTPException422(detail="Output path not specified.")
    return node.storage


@data_file_router.get("/")
def get_node_storage(
    storage: Annotated[DataFileStorage, Depends(_get_storage_instance)],
) -> str:
    return str(storage.path)


@data_file_router.get("/content")
def get_node_storage_content(
    *,
    load_type: StorageLoadType = StorageLoadType.Full,
    storage: Annotated[DataFileStorage, Depends(_get_storage_instance)],
) -> Optional[DocumentType]:
    storage.load(load_type)
    return storage.data



import json
from base64 import b64encode
from collections.abc import Mapping
from enum import IntEnum
from pathlib import Path
from typing import Any, Optional

from qualibrate_config.models import QualibrateConfig

from qualibrate_app.api.core.domain.bases.i_dump import IDump
from qualibrate_app.api.core.models.storage import Storage as StorageModel
from qualibrate_app.api.exceptions.classes.storage import QFileNotFoundException
from qualibrate_app.api.exceptions.classes.values import QValueException

__all__ = ["DataFileStorage", "StorageLoadType"]


class StorageLoadType(IntEnum):
    Empty = 0
    Full = 1


class DataFileStorage(IDump):
    data_file_name = "data.json"

    def __init__(self, path: Path, settings: QualibrateConfig):
        if not path.is_dir():
            rel_path = path.relative_to(settings.storage.location)
            raise QFileNotFoundException(f"{rel_path} does not exist.")
        self._path = path
        self._load_type = StorageLoadType.Empty
        self._data: Optional[Mapping[str, Any]] = None
        self._settings = settings

    @property
    def path(self) -> Path:
        return self._path.relative_to(self._settings.storage.location)

    @property
    def data(self) -> Optional[Mapping[str, Any]]:
        return self._data

    @property
    def load_type(self) -> StorageLoadType:
        return self._load_type

    def load(self, load_type: StorageLoadType) -> None:
        if self.load_type >= load_type:
            return
        if load_type != StorageLoadType.Full:
            return
        self._parse_data()

    def exists(self) -> bool:
        return self._path.exists()

    def list_elements(self) -> list[str]:  # iterdir isn't recursive
        return list(path.name for path in self._path.iterdir())

    def list_typed_elements(self, pattern: str) -> list[str]:
        return list(path.name for path in self._path.glob(pattern))

    def _recursively_parse_data(
        self, content: dict[str, Any]
    ) -> dict[str, Any]:
        for key, value in content.items():
            if isinstance(value, str) and Path(value).suffix == ".png":
                img_path = (self._path / value).resolve()
                if (
                    not img_path.is_relative_to(self._path)
                    or not img_path.is_file()
                ):
                    continue
                img_data = img_path.read_bytes()
                # TODO: dynamic compute MIME type
                content[key] = {
                    value: (
                        "data:image/png;base64,"
                        f"{b64encode(img_data).decode('utf-8')}"
                    )
                }
            if isinstance(value, dict):
                content[key] = self._recursively_parse_data(value)
        return content

    def _parse_data(self) -> None:
        data_file = self._path / self.__class__.data_file_name
        if not data_file.is_file():
            return None
        with data_file.open("r") as file:
            try:
                content = json.load(file)
            except json.JSONDecodeError as ex:
                raise QValueException("Unexpected data format.") from ex
        if not isinstance(content, Mapping):
            raise QValueException("Unexpected data format.")
        self._data = self._recursively_parse_data(dict(content))
        self._load_type = StorageLoadType.Full

    def dump(self) -> StorageModel:
        return StorageModel(path=self.path, data=self.data)



from starlette import status
from starlette.middleware.base import (
    BaseHTTPMiddleware,
    RequestResponseEndpoint,
)
from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from starlette.types import ASGIApp

from qualibrate_app.api.exceptions.classes.base import QualibrateException


class QualibrateCatchExcMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp):
        super().__init__(app)

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        try:
            return await call_next(request)
        except QualibrateException as exc:
            return JSONResponse(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                content={"detail": str(exc)},
                headers={"X-Exception-Type": exc.__class__.__name__},
            )



from collections.abc import Sequence
from typing import Annotated, Union

from fastapi import Query

from qualibrate_app.api.core.types import AllowedSearchKeys
from qualibrate_app.api.core.utils.request_utils import HTTPException422


def check_path_item(item: str) -> bool:
    return len(item) > 0 and all(c in AllowedSearchKeys for c in item)


def check_path(value: str) -> str:
    """
    Check that all symbols in path in ascii + digits + `_-`

    Raises:
        `InvalidMongoPath` if path is invalid.

    Returns:
        Passed path
    """
    if not all(check_path_item(subpath) for subpath in value.split(".")):
        raise HTTPException422(detail=f"Invalid mongo search path '{value}'")
    return value


def get_search_path(
    data_path: Annotated[str, Query(description="Path to search")],
) -> Sequence[Union[str, int]]:
    """Generate list of search subpaths from search path.

    Raises:
        `InvalidMongoPath` if path is invalid.

    Returns:
        List of item subpaths
    """
    return [
        int(item) if item.isnumeric() else item
        for item in check_path(data_path).split(".")
    ]



import contextlib
import json
import logging
from collections.abc import Mapping, MutableMapping, Sequence
from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    Optional,
    Union,
    cast,
)
from urllib.parse import urljoin

import jsonpatch
import jsonpointer
import requests
from qualibrate_config.models import QualibrateConfig
from requests import JSONDecodeError as RequestsJSONDecodeError

from qualibrate_app.api.core.domain.bases.snapshot import (
    SnapshotBase,
    SnapshotLoadType,
)
from qualibrate_app.api.core.domain.local_storage._id_to_local_path import (
    IdToLocalPath,
)
from qualibrate_app.api.core.domain.local_storage.utils import (
    snapshot_content as snapshot_content_utils,
)
from qualibrate_app.api.core.domain.local_storage.utils.node_utils import (
    find_latest_node_id,
    find_n_latest_nodes_ids,
)
from qualibrate_app.api.core.schemas.state_updates import StateUpdates
from qualibrate_app.api.core.types import (
    DocumentSequenceType,
    DocumentType,
    IdType,
)
from qualibrate_app.api.core.utils.find_utils import get_subpath_value
from qualibrate_app.api.core.utils.path.node import NodePath
from qualibrate_app.api.core.utils.snapshots_compare import jsonpatch_to_mapping
from qualibrate_app.api.core.utils.types_parsing import TYPE_TO_STR
from qualibrate_app.api.exceptions.classes.storage import (
    QFileNotFoundException,
    QPathException,
)
from qualibrate_app.api.exceptions.classes.values import QValueException

__all__ = ["SnapshotLocalStorage"]

logger = logging.getLogger(__name__)


class SnapshotLocalStorage(SnapshotBase):
    """
    A class for managing local storage of snapshots, inheriting from
    `SnapshotBase`.

    Args:
        id: The identifier of the snapshot.
        content: The content of the snapshot. Defaults to None.
        snapshot_loader: Function to load snapshot content.
            Defaults to `_default_snapshot_content_loader`.
        snapshot_updater: Function to update snapshot content.
            Defaults to `_default_snapshot_content_updater`.
        settings: The application settings for Qualibrate.

    Notes:
        Expected structure of content root:
        - base_path
            - %Y-%m-%d
                - #{idx}_{name}_%H%M%S  # node
                    - data.json    # outputs
                    - state.json   # QuAM state
            - %Y-%m-%d
            ...
    """

    def __init__(
        self,
        id: IdType,
        content: Optional[DocumentType] = None,
        snapshot_loader: snapshot_content_utils.SnapshotContentLoaderType = (
            snapshot_content_utils.default_snapshot_content_loader
        ),
        snapshot_updater: snapshot_content_utils.SnapshotContentUpdaterType = (
            snapshot_content_utils.default_snapshot_content_updater
        ),
        *,
        settings: QualibrateConfig,
    ):
        """Initializes the SnapshotLocalStorage instance."""
        super().__init__(id=id, content=content, settings=settings)
        self._snapshot_loader = snapshot_loader
        self._snapshot_updater = snapshot_updater
        self._node_path: Optional[NodePath] = None

    @property
    def node_path(self) -> NodePath:
        """
        Returns the path to the node.

        Returns:
            The path to the node.
        """
        if self._node_path is None:
            self._node_path = IdToLocalPath().get_or_raise(
                self._settings.project,
                self._id,
                self._settings.storage.location,
            )
        return self._node_path

    def load(self, load_type: SnapshotLoadType) -> None:
        """
        Loads snapshot content based on the specified load type.

        Args:
            load_type: The type of content to load.
        """
        if load_type <= self._load_type:
            return None
        content = self._snapshot_loader(
            self.node_path, load_type, self._settings
        )
        self.content.update(content)
        self._load_type = load_type

    @property
    def created_at(self) -> Optional[datetime]:
        """
        Returns the creation date of the snapshot.

        Returns:
            The creation date or None if not available.
        """
        return self.content.get("created_at")

    @property
    def parents(self) -> Optional[list[IdType]]:
        """
        Returns the list of parent snapshot IDs.

        Returns:
            The parent IDs or None if not available.
        """
        return self.content.get("parents")

    def search(
        self, search_path: Sequence[Union[str, int]], load: bool = False
    ) -> Optional[DocumentSequenceType]:
        """
        Searches for a value in the snapshot data at a specified path.

        Args:
            search_path: The path to search.
            load: Whether to load the data before searching. Defaults to False.

        Returns:
            The found value or None if not found.
        """
        if load:
            self.load(SnapshotLoadType.Data)
        if self.data is None or "quam" not in self.data:
            return None
        # TODO: update logic; not use quam directly
        return get_subpath_value(self.data["quam"], search_path)

    def get_latest_snapshots(
        self, page: int = 1, per_page: int = 50, reverse: bool = False
    ) -> tuple[int, Sequence[SnapshotBase]]:
        """
        Retrieves the latest snapshots. First item in sequence is current.

        Args:
            page: The page number. Defaults to 1.
            per_page: The number of snapshots per page. Defaults to 50.
            reverse: Whether to reverse the order. Defaults to False.

        Returns:
            Total number of snapshots and a sequence of the latest snapshots.
        """
        storage_location = self._settings.storage.location
        total = find_latest_node_id(storage_location)
        self.load(SnapshotLoadType.Metadata)
        if page == 1 and per_page == 1:
            return total, [self]
        ids = find_n_latest_nodes_ids(
            storage_location,
            page,
            per_page,
            self._settings.project,
            max_node_id=(self.id or total) - 1,
        )
        snapshots = [
            SnapshotLocalStorage(id, settings=self._settings) for id in ids
        ]
        for snapshot in snapshots:
            with contextlib.suppress(OSError):
                snapshot.load(SnapshotLoadType.Metadata)
        return total, [self, *snapshots]

    def compare_by_id(
        self, other_snapshot_id: int
    ) -> Mapping[str, Mapping[str, Any]]:
        """
        Compares the current snapshot with another snapshot by ID.

        Args:
            other_snapshot_id: The ID of the other snapshot.

        Returns:
            The comparison result as a mapping.

        Raises:
            QValueException: If comparing with the same snapshot ID or if data
                cannot be loaded.
        """
        if self.id == other_snapshot_id:
            raise QValueException("Can't compare snapshots with same id")
        self.load(SnapshotLoadType.Data)
        # TODO: update logic; not use quam directly
        this_data = (self.data or {}).get("quam")
        if this_data is None:
            raise QValueException(f"Can't load data of snapshot {self._id}")
        other_snapshot = SnapshotLocalStorage(
            other_snapshot_id, settings=self._settings
        )
        other_snapshot.load(SnapshotLoadType.Data)
        # TODO: update logic; not use quam directly
        other_data = (other_snapshot.data or {}).get("quam")
        if other_data is None:
            raise QValueException(
                f"Can't load data of snapshot {other_snapshot_id}"
            )
        return jsonpatch_to_mapping(
            this_data, jsonpatch.make_patch(dict(this_data), dict(other_data))
        )

    @staticmethod
    def _conversion_type_from_value(value: Any) -> Mapping[str, Any]:
        """
        Determines the type of given value for JSON schema conversion.

        Args:
            value: The value to analyze.

        Returns:
            A mapping indicating the type.
        """
        if isinstance(value, list):
            if len(value) == 0:
                return {"type": "array"}
            item_type = TYPE_TO_STR.get(type(value[0]))
            if item_type is None:
                return {"type": "array"}
            return {"type": "array", "items": {"type": item_type}}
        item_type = TYPE_TO_STR.get(type(value))
        if item_type is None:
            return {"type": "null"}
        return {"type": item_type}

    def get_state_updates_from_runner(
        self,
        **kwargs: Any,
    ) -> Optional[StateUpdates]:
        """
        Retrieves state updates from the runner.

        Args:
            **kwargs: Additional arguments such as cookies for the request.

        Returns:
            The state updates or None if retrieval fails.
        """
        if self._settings.runner is None:
            logger.warning("Runner config is not set")
            return None
        try:
            cookies = cast(
                Optional[MutableMapping[str, str]], kwargs.get("cookies")
            )
            last_run_response = requests.get(
                urljoin(self._settings.runner.address_with_root, "last_run/"),
                cookies=cookies,
            )
        except requests.exceptions.ConnectionError:
            return None
        if last_run_response.status_code != 200:
            return None
        try:
            data = last_run_response.json()
        except RequestsJSONDecodeError:
            return None
        if data is None:
            return None
        return StateUpdates(state_updates=data.get("state_updates", {}))

    def _extract_state_update_type_from_runner(
        self,
        path: str,
        state_updates: Optional[StateUpdates] = None,
        **kwargs: Any,
    ) -> Optional[Mapping[str, Any]]:
        """
        Extracts the state update type for a specific path from the runner.

        Args:
            path: The path to extract the state update type from.
            state_updates: Pre-fetched state updates. Defaults to None.
            **kwargs: Additional arguments for retrieving state updates.

        Returns:
            The type mapping for the state update, or None if not found.
        """
        if state_updates is None:
            state_updates = self.get_state_updates_from_runner(**kwargs)
        if state_updates is None or path not in state_updates.state_updates:
            return None
        new_state = state_updates.state_updates[path].new
        return self._conversion_type_from_value(new_state)

    def extract_state_update_types_from_runner(
        self,
        paths: Sequence[str],
        state_updates: Optional[StateUpdates] = None,
        **kwargs: Any,
    ) -> Mapping[str, Optional[Mapping[str, Any]]]:
        """
        Extracts state update types for multiple paths from the runner.

        Args:
            paths: A sequence of paths to extract state update types for.
            state_updates: Pre-fetched state updates. Defaults to None.
            **kwargs: Additional arguments for retrieving state updates.

        Returns:
            A mapping of paths to their respective state update types.
        """
        if state_updates is None:
            state_updates = self.get_state_updates_from_runner(**kwargs)
        if state_updates is None:
            return {}
        return {
            path: self._extract_state_update_type_from_runner(
                path, state_updates
            )
            for path in paths
        }

    def get_quam_state(
        self,
    ) -> Optional[Mapping[str, Any]]:
        """
        Retrieves the QuAM state from the snapshot.

        Returns:
            The QuAM state data or None if not available.
        """
        try:
            quam_state_file: Path = self.node_path / "quam_state.json"
        except QFileNotFoundException:
            return None
        # TODO: fix quam state resolved (should check data property)
        if not quam_state_file.is_file():
            quam_state_file = self.node_path / "state.json"
        if not quam_state_file.is_file():
            return None
        try:
            return cast(
                Mapping[str, Any], json.loads(quam_state_file.read_text())
            )
        except json.JSONDecodeError:
            return None

    def _extract_state_update_type_from_quam_state(
        self, path: str, quam_state: Optional[Mapping[str, Any]] = None
    ) -> Optional[Mapping[str, Any]]:
        """
        Extracts the state update type for a specific path from the QuAM state.

        Args:
            path: The path to extract the state update type from.
            quam_state: Pre-fetched QuAM state. Defaults to None.

        Returns:
            The type mapping for the state update, or None if not found.
        """
        if quam_state is None:
            quam_state = self.get_quam_state()
        if quam_state is None:
            return None
        quam_item = jsonpointer.resolve_pointer(quam_state, path[1:], object)
        if quam_item is object:
            return None
        return self._conversion_type_from_value(quam_item)

    def extract_state_update_types_from_quam_state(
        self,
        paths: Sequence[str],
        quam_state: Optional[Mapping[str, Any]] = None,
    ) -> Optional[Mapping[str, Any]]:
        """
        Extracts state update types for multiple paths from the QuAM state.

        Args:
            paths: A sequence of paths to extract state update types for.
            quam_state: Pre-fetched QuAM state. Defaults to None.

        Returns:
            A mapping of paths to their respective state update types.
        """
        if quam_state is None:
            quam_state = self.get_quam_state()
        if quam_state is None:
            return None
        return {
            path: self._extract_state_update_type_from_quam_state(
                path, quam_state
            )
            for path in paths
        }

    def _saved_data_quam_path(self, path: str) -> str:
        parts = path.split("/", maxsplit=1)
        if len(parts) == 1:
            return path
        return "/".join([parts[0], "quam", parts[1]])

    def extract_state_update_type(
        self,
        path: str,
        **kwargs: Mapping[str, Any],
    ) -> Optional[Mapping[str, Any]]:
        """
        Extracts the state update type for a specific path from either the
        runner or QuAM state.

        Args:
            path: The path to extract the state update type from.
            **kwargs: Additional arguments for retrieving state updates.

        Returns:
            The type mapping for the state update, or None if not found.
        """
        _type = self._extract_state_update_type_from_runner(
            path, None, **kwargs
        )
        if _type is not None:
            return _type
        return self._extract_state_update_type_from_quam_state(path)

    def extract_state_update_types(
        self,
        paths: Sequence[str],
        **kwargs: Mapping[str, Any],
    ) -> Mapping[str, Optional[Mapping[str, Any]]]:
        """
        Extracts state update types for multiple paths from either the runner
        or QuAM state.

        Args:
            paths: A sequence of paths to extract state update types for.
            **kwargs: Additional arguments for retrieving state updates.

        Returns:
            A mapping of paths to their respective state update types.
        """
        types = self.extract_state_update_types_from_runner(
            paths, None, **kwargs
        )
        if len(types):
            return types
        return self.extract_state_update_types_from_quam_state(paths) or {}

    def update_entry(self, updates: Mapping[str, Any]) -> bool:
        """
        Updates the snapshot data with specified updates.

        Args:
            updates: A mapping of paths to new values for the update.

        Returns:
            True if the update was successful, False otherwise.

        Raises:
            QPathException: If an unknown path is encountered during the update.
        """
        if self.load_type < SnapshotLoadType.Data:
            self.load(SnapshotLoadType.Data)
        data = self.data
        if data is None or not isinstance(data.get("quam"), Mapping):
            return False
        # override update paths with quam prefix
        updates = {self._saved_data_quam_path(k): v for k, v in updates.items()}

        path_values = {
            path: jsonpointer.resolve_pointer(data, path[1:], None)
            for path in updates
        }
        replace_updates = filter(
            lambda k: path_values[k] is not None, updates.keys()
        )
        add_updates = filter(lambda k: path_values[k] is None, updates.keys())
        replace_patch_operations = [
            {
                "op": "replace",
                "path": path[1:],
                "value": updates[path],
                "old": path_values[path],
            }
            for path in replace_updates
        ]
        add_patch_operations = [
            {"op": "add", "path": path[1:], "value": updates[path]}
            for path in add_updates
        ]
        patch_operations = add_patch_operations + replace_patch_operations
        patch = jsonpatch.JsonPatch(patch_operations)
        try:
            new_data = patch.apply(dict(data))
            res = self._snapshot_updater(
                self.node_path, new_data, patch_operations, self._settings
            )
            return res
        except jsonpatch.JsonPatchException as ex:
            raise QPathException("Unknown path to update") from ex
        except OSError:
            return False



from collections.abc import Sequence
from typing import Any, Optional, Union

from qualibrate_app.api.core.domain.bases.branch import BranchBase
from qualibrate_app.api.core.domain.bases.node import NodeBase
from qualibrate_app.api.core.domain.bases.root import RootBase
from qualibrate_app.api.core.domain.bases.snapshot import SnapshotBase
from qualibrate_app.api.core.domain.local_storage.branch import (
    BranchLocalStorage,
)
from qualibrate_app.api.core.domain.local_storage.node import NodeLocalStorage
from qualibrate_app.api.core.domain.local_storage.snapshot import (
    SnapshotLocalStorage,
)
from qualibrate_app.api.core.domain.local_storage.utils.node_utils import (
    find_n_latest_nodes_ids,
)
from qualibrate_app.api.core.types import IdType
from qualibrate_app.api.exceptions.classes.storage import QFileNotFoundException

__all__ = ["RootLocalStorage"]


class RootLocalStorage(RootBase):
    def get_branch(self, branch_name: str) -> BranchBase:
        return BranchLocalStorage(branch_name, settings=self._settings)

    def _get_latest_node_id(self, error_msg: str) -> IdType:
        id = next(
            find_n_latest_nodes_ids(
                self._settings.storage.location,
                1,
                1,
                self._settings.project,
            ),
            None,
        )
        if id is None:
            raise QFileNotFoundException(f"There is no {error_msg}")
        return id

    def get_snapshot(self, id: Optional[IdType] = None) -> SnapshotBase:
        if id is None:
            id = self._get_latest_node_id("snapshot")
        return SnapshotLocalStorage(id, settings=self._settings)

    def get_node(self, id: Optional[IdType] = None) -> NodeBase:
        if id is None:
            id = self._get_latest_node_id("node")
        return NodeLocalStorage(id, settings=self._settings)

    def get_latest_snapshots(
        self,
        page: int = 1,
        per_page: int = 50,
        reverse: bool = False,
    ) -> tuple[int, Sequence[SnapshotBase]]:
        return BranchLocalStorage(
            "main", settings=self._settings
        ).get_latest_snapshots(page, per_page, reverse)

    def get_latest_nodes(
        self,
        page: int = 1,
        per_page: int = 50,
        reverse: bool = False,
    ) -> tuple[int, Sequence[NodeBase]]:
        return BranchLocalStorage(
            "main", settings=self._settings
        ).get_latest_nodes(page, per_page, reverse)

    def search_snapshot(
        self, snapshot_id: IdType, data_path: Sequence[Union[str, int]]
    ) -> Any:
        return self.get_snapshot(snapshot_id).search(data_path, load=True)



import uvicorn
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from starlette.middleware.cors import CORSMiddleware

from qualibrate_app.api.__main__ import api_router
from qualibrate_app.api.exceptions.middleware import (
    QualibrateCatchExcMiddleware,
)
from qualibrate_app.config.resolvers import (
    get_config_path,
    get_default_static_files_path,
    get_settings,
)

try:
    from json_timeline_database.app import app as json_timeline_db_app
except ImportError:
    json_timeline_db_app = None


app = FastAPI(
    title="Qualibrate",
    openapi_url="/app_openapi.json",
    docs_url="/app_docs",
)
_settings = get_settings(get_config_path())

origins = [
    "http://localhost:8002",
    "http://localhost:8001",
    "http://127.0.0.1:8002",
    "http://127.0.0.1:8001",
]

app.add_middleware(QualibrateCatchExcMiddleware)
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(api_router, prefix="/api")

static_files_path = (
    _settings.app.static_site_files
    if (
        _settings is not None
        and _settings.app is not None
        and _settings.app.static_site_files is not None
    )
    else get_default_static_files_path()
)
if static_files_path is None or not static_files_path.is_dir():
    raise RuntimeError(
        "No static files found in config.toml or default location"
    )
# Directory should exist
app.mount(
    "/",
    StaticFiles(directory=static_files_path, html=True),
    name="static",
)


def main(port: int, reload: bool) -> None:
    uvicorn.run("qualibrate_app.app:app", port=port, reload=reload)


if __name__ == "__main__":
    main(port=8001, reload=False)



from functools import wraps
from typing import Any, Callable, TypeVar, cast

from qualibrate_app.api.exceptions.classes.types import QInvalidIdTypeException

Function = TypeVar("Function", bound=Callable[..., Any])


def id_type_str(func: Function) -> Function:
    @wraps(func)
    def wrapper(*args, **kwargs):  # type: ignore
        err = QInvalidIdTypeException("id should be str")
        if "id" in kwargs and not isinstance(kwargs["id"], str):
            raise err
        elif len(args) > 0 and len(args) > 1 and not isinstance(args[1], str):
            # temporary only for methods
            raise err
        return func(*args, **kwargs)

    return cast(Function, wrapper)



import os
import warnings
from functools import lru_cache
from pathlib import Path
from typing import Annotated, Optional

from fastapi import Depends
from qualibrate_config.models import (
    QualibrateConfig,
)
from qualibrate_config.resolvers import (
    get_qualibrate_config,
    get_qualibrate_config_path,
)

from qualibrate_app.config.vars import (
    CONFIG_PATH_ENV_NAME,
)

__all__ = [
    "get_default_static_files_path",
    "get_config_path",
    "get_settings",
    "get_quam_state_path",
]


def get_default_static_files_path() -> Optional[Path]:
    import sys

    module_file = sys.modules["qualibrate_app"].__file__
    if module_file is None:
        return None
    module_path = Path(module_file)
    return module_path.parents[1] / "qualibrate_static"


@lru_cache
def get_config_path() -> Path:
    path = os.environ.get(CONFIG_PATH_ENV_NAME)
    if path is not None:
        return Path(path)
    return get_qualibrate_config_path()


@lru_cache
def get_settings(
    config_path: Annotated[Path, Depends(get_config_path)],
) -> QualibrateConfig:
    return get_qualibrate_config(config_path)


@lru_cache
def get_quam_state_path(
    settings: Annotated[QualibrateConfig, Depends(get_settings)],
) -> Optional[Path]:
    root = settings.__class__._root
    if root is None:
        return None
    quam_state_path = root._raw_dict.get("quam", {}).get("state_path")
    if quam_state_path is not None:
        return Path(quam_state_path)
    active_machine_path = root._raw_dict.get("active_machine", {}).get("path")
    if active_machine_path is None:
        return None
    warnings.warn(
        (
            'The config entry "active_machine.path" has been deprecated in '
            'favor of "quam.state_path". Please update the qualibrate config '
            "(~/.qualibrate/config.toml) accordingly."
        ),
        DeprecationWarning,
        stacklevel=2,
    )
    return Path(active_machine_path)



from typing import Annotated, Any, Optional

from pydantic import AwareDatetime, BaseModel, ConfigDict, Field, computed_field

from qualibrate_app.api.core.models.base import ModelWithIdCreatedAt
from qualibrate_app.api.core.types import IdType


class SimplifiedSnapshot(ModelWithIdCreatedAt):
    parents: list[IdType]


class SnapshotMetadata(BaseModel):
    model_config = ConfigDict(extra="allow")

    status: Optional[str] = None
    description: Optional[str] = None
    run_start: Optional[AwareDatetime] = None
    run_end: Optional[AwareDatetime] = None

    @computed_field
    def run_duration(self) -> Optional[float]:
        if self.run_start is None or self.run_end is None:
            return None
        return round((self.run_end - self.run_start).total_seconds(), 3)


class SimplifiedSnapshotWithMetadata(SimplifiedSnapshot):
    metadata: SnapshotMetadata


class SnapshotData(BaseModel):
    model_config = ConfigDict(extra="allow")

    quam: Optional[dict[str, Any]] = None
    parameters: Optional[dict[str, Any]] = None
    outcomes: Optional[dict[str, Any]] = None


class Snapshot(SimplifiedSnapshot):
    metadata: Annotated[
        SnapshotMetadata, Field(default_factory=SnapshotMetadata)
    ]
    data: Optional[SnapshotData] = None



from collections.abc import Sequence
from datetime import datetime
from typing import Optional

from qualibrate_config.models import QualibrateConfig

from qualibrate_app.api.core.domain.bases.branch import (
    BranchBase,
    BranchLoadType,
)
from qualibrate_app.api.core.domain.bases.node import NodeBase, NodeLoadType
from qualibrate_app.api.core.domain.bases.snapshot import (
    SnapshotBase,
    SnapshotLoadType,
)
from qualibrate_app.api.core.domain.local_storage.node import NodeLocalStorage
from qualibrate_app.api.core.domain.local_storage.snapshot import (
    SnapshotLocalStorage,
)
from qualibrate_app.api.core.domain.local_storage.utils.node_utils import (
    find_latest_node_id,
    find_n_latest_nodes_ids,
)
from qualibrate_app.api.core.models.branch import Branch as BranchModel
from qualibrate_app.api.core.types import DocumentType, IdType
from qualibrate_app.api.exceptions.classes.storage import QFileNotFoundException

__all__ = ["BranchLocalStorage"]


class BranchLocalStorage(BranchBase):
    def __init__(
        self,
        name: str,
        content: Optional[DocumentType] = None,
        *,
        settings: QualibrateConfig,
    ):
        # Temporary branch name has no effect
        super().__init__(name, content, settings=settings)

    @property
    def created_at(self) -> datetime:
        return datetime.fromtimestamp(
            self._settings.storage.location.stat().st_mtime
        ).astimezone()

    def load(self, load_type: BranchLoadType) -> None:
        pass

    def _get_latest_node_id(self, error_msg: str) -> IdType:
        id = next(
            find_n_latest_nodes_ids(
                self._settings.storage.location,
                1,
                1,
                self._settings.project,
            ),
            None,
        )
        if id is None:
            raise QFileNotFoundException(f"There is no {error_msg}")
        return id

    def get_snapshot(self, id: Optional[IdType] = None) -> SnapshotBase:
        if id is None:
            id = self._get_latest_node_id("snapshots")
        return SnapshotLocalStorage(id, settings=self._settings)

    def get_node(self, id: Optional[IdType] = None) -> NodeBase:
        if id is None:
            id = self._get_latest_node_id("nodes")
        return NodeLocalStorage(id, settings=self._settings)

    def get_latest_snapshots(
        self,
        page: int = 0,
        per_page: int = 50,
        reverse: bool = False,
    ) -> tuple[int, Sequence[SnapshotBase]]:
        # TODO: use reverse
        storage_location = self._settings.storage.location
        ids = find_n_latest_nodes_ids(
            storage_location,
            page,
            per_page,
            self._settings.project,
        )
        snapshots = [
            SnapshotLocalStorage(id, settings=self._settings) for id in ids
        ]
        for snapshot in snapshots:
            snapshot.load(SnapshotLoadType.Metadata)
        total = find_latest_node_id(storage_location)
        return total, snapshots

    def get_latest_nodes(
        self,
        page: int = 0,
        per_page: int = 50,
        reverse: bool = False,
    ) -> tuple[int, Sequence[NodeBase]]:
        # TODO: use reverse
        storage_location = self._settings.storage.location
        ids = find_n_latest_nodes_ids(
            storage_location,
            page,
            per_page,
            self._settings.project,
        )
        nodes = [NodeLocalStorage(id, settings=self._settings) for id in ids]
        for node in nodes:
            node.load(NodeLoadType.Full)
        total = find_latest_node_id(storage_location)
        return total, nodes

    def dump(self) -> BranchModel:
        return BranchModel(
            id=1,
            created_at=self.created_at,
            name=self._name,
            snapshot_id=-1,
        )



import os
import sys
from collections.abc import Mapping
from pathlib import Path
from typing import Any

import click
from click.core import ParameterSource
from qualibrate_config.resolvers import get_qualibrate_config_path

if sys.version_info[:2] < (3, 11):
    import tomli as tomllib
else:
    import tomllib


# __all__ = ["config_command"]


def not_default(ctx: click.Context, arg_key: str) -> bool:
    return ctx.get_parameter_source(arg_key) in (
        ParameterSource.COMMANDLINE,
        ParameterSource.ENVIRONMENT,
    )


def get_config(config_path: Path) -> tuple[dict[str, Any], Path]:
    """Returns config and path to file"""
    config_file = get_qualibrate_config_path()
    if config_file.is_file():
        return tomllib.loads(config_file.read_text()), config_path
    return {}, config_file


def _config_from_sources(
    ctx: click.Context, from_file: dict[str, Any]
) -> dict[str, Any]:
    qualibrate_mapping = {k: k for k in ("static_site_files",)}
    timeline_db_mapping = {
        "timeline_db_address": "address",
        "timeline_db_timeout": "timeout",
    }
    runner_mapping = {
        "runner_address": "address",
        "runner_timeout": "timeout",
    }
    for arg_key, arg_value in ctx.params.items():
        not_default_arg = not_default(ctx, arg_key)
        if arg_key in qualibrate_mapping:
            if not_default_arg or qualibrate_mapping[arg_key] not in from_file:
                from_file[qualibrate_mapping[arg_key]] = arg_value
        elif arg_key in timeline_db_mapping:
            if not_default_arg or (
                timeline_db_mapping[arg_key] not in from_file["timeline_db"]
            ):
                from_file["timeline_db"][timeline_db_mapping[arg_key]] = (
                    arg_value
                )
        elif arg_key in runner_mapping and (
            not_default_arg
            or (runner_mapping[arg_key] not in from_file["runner"])
        ):
            from_file["runner"][runner_mapping[arg_key]] = arg_value
    return from_file


def _print_config(data: Mapping[str, Any], depth: int = 0) -> None:
    try:
        max_key_len = max(map(len, map(str, data.keys())))
    except ValueError:
        max_key_len = 1
    click.echo(
        os.linesep.join(
            f"{' ' * 4 * depth}{f'{k} :':<{max_key_len + 3}} {v}"
            for k, v in data.items()
            if not isinstance(v, Mapping)
        )
    )
    mappings = filter(lambda x: isinstance(x[1], Mapping), data.items())
    for mapping_k, mapping_v in mappings:
        click.echo(f"{' ' * 4 * depth}{mapping_k} :")
        _print_config(mapping_v, depth + 1)


def _confirm(
    config_file: Path,
    qs_data: Mapping[str, Any],
    am_data: Mapping[str, Any],
    qas_data: Mapping[str, Any],
) -> None:
    click.echo(f"Config file path: {config_file}")
    click.echo(click.style("Generated config:", bold=True))
    _print_config(
        {
            "Qualibrate": qs_data,
            "Qualibrate app": qas_data,
            "Active machine": am_data,
        }
    )
    confirmed = click.confirm("Do you confirm config?", default=True)
    if not confirmed:
        click.echo(
            click.style(
                (
                    "The configuration has not been confirmed. "
                    "Rerun config script."
                ),
                fg="yellow",
            )
        )
        exit(1)


# def write_config(
#     config_file: Path,
#     common_config: dict[str, Any],
#     qss: QualibrateSettingsSetup,
#     ams: ActiveMachineSettingsSetup,
#     qass: QualibrateAppSettingsSetup,
#     confirm: bool = True,
# ) -> None:
#     qs_exported_data = qss.model_dump(exclude_none=True)
#     am_exported_data = ams.model_dump(exclude_none=True)
#     qas_exported_data = qass.model_dump()
#     if confirm:
#         _confirm(
#             config_file, qs_exported_data, am_exported_data, qas_exported_data
#         )
#     storage_location = cast(Path, qss.storage.location)
#     storage_location.mkdir(parents=True, exist_ok=True)
#     if qss.project:
#         project_path = storage_location / qss.project
#         project_path.mkdir(parents=True, exist_ok=True)
#     if qss.log_folder:
#         cast(Path, qss.log_folder).mkdir(parents=True, exist_ok=True)
#     if not config_file.parent.exists():
#         config_file.parent.mkdir(parents=True)
#     common_config[CONFIG_KEY] = qas_exported_data
#     common_config[ACTIVE_MACHINE_CONFIG_KEY] = am_exported_data
#     common_config[QUALIBRATE_CONFIG_KEY] = qs_exported_data
#
#     with config_file.open("wb") as f_out:
#         tomli_w.dump(common_config, f_out)
#
#
# @click.command(name="config")
# @click.option(
#     "--config-path",
#     type=click.Path(
#         exists=False,
#         path_type=Path,
#     ),
#     default=QUALIBRATE_PATH / DEFAULT_CONFIG_FILENAME,
#     show_default=True,
# )
# @click.option(
#     "--auto-accept",
#     type=bool,
#     is_flag=True,
#     default=False,
#     show_default=True,
# )
# @click.option(
#     "--overwrite",
#     type=bool,
#     default=False,
#     is_flag=True,
#     help="Ignore existing config and force overwrite values",
# )
# @click.option(
#     "--static-site-files",
#     type=click.Path(
#         exists=True,
#         file_okay=False,
#         dir_okay=True,
#         resolve_path=True,
#         path_type=Path,
#     ),
#     default=Path(__file__).parents[2] / "qualibrate_static",
# )
# @click.option(
#     "--storage-type",
#     type=click.Choice([t.value for t in StorageType]),
#     default=StorageType.local_storage.value,
#     show_default=True,
# )
# @click.option(
#     "--user-storage",
#     type=click.Path(
#         exists=False,
#         resolve_path=True,
#         path_type=Path,
#     ),
#     default=QUALIBRATE_PATH / "user_storage",
#     help="Path to user storage directory with qualibrate data.",
#     show_default=True,
# )
# @click.option(
#     "--project",
#     type=str,
#     required=False,
# )
# @click.option(
#     "--timeline-db-address",
#     type=str,  # TODO: add type check for addr
#     default="http://localhost:8000/",
#     show_default=True,
# )
# @click.option(
#     "--timeline-db-timeout",
#     type=float,
#     default=1.0,
#     show_default=True,
# )
# @click.option(
#     "--runner-address",
#     type=str,  # TODO: add type check for addr
#     default="http://localhost:8001/execution/",
#     show_default=True,
# )
# @click.option(
#     "--runner-timeout",
#     type=float,
#     default=1.0,
#     show_default=True,
# )
# @click.pass_context
# def config_command(
#     ctx: click.Context,
#     config_path: Path,
#     auto_accept: bool,
#     overwrite: bool,
#     static_site_files: Path,
#     storage_type: StorageType,
#     log_path: Path,
#     user_storage: Path,
#     project: Optional[str],
#     timeline_db_address: str,
#     timeline_db_timeout: float,
#     runner_address: str,
#     runner_timeout: float,
# ) -> None:
#     common_config, config_file = get_config(config_path)
#     qualibrate_app_config = (
#         common_config.get(CONFIG_KEY, {}) if not overwrite else {}
#     )
#     qapp_subconfigs = ("timeline_db", "runner")
#     for subconfig in qapp_subconfigs:
#         if subconfig not in qualibrate_app_config:
#             qualibrate_app_config[subconfig] = {}
#
#     qualibrate_app_config = _config_from_sources(ctx, qualibrate_app_config)
#     (
#         qualibrate_config,
#         active_machine_config,
#         qapp_config,
#     ) = check_config_pre_v1_and_update(common_config, qualibrate_app_config)
#     qss = get_config_model_or_print_error(
#         qualibrate_config, QualibrateSettingsSetup, QUALIBRATE_CONFIG_KEY
#     )
#     ams = get_config_model_or_print_error(
#         active_machine_config,
#         ActiveMachineSettingsSetup,
#         ACTIVE_MACHINE_CONFIG_KEY,
#     )
#     qass = get_config_model_or_print_error(
#         qapp_config, QualibrateAppSettingsSetup, CONFIG_KEY
#     )
#     if qss is None or ams is None or qass is None:
#         return
#     write_config(
#         config_file,
#         common_config,
#         qss,
#         ams,
#         qass,
#         confirm=not auto_accept,
#     )
